package arm

import (
	"git.urbach.dev/cli/q/src/cpu"
	"git.urbach.dev/cli/q/src/sizeof"
)

// MoveRegisterRegister copies a register to another register.
func MoveRegisterRegister(destination cpu.Register, source cpu.Register) uint32 {
	if source == SP || destination == SP {
		return AddRegisterNumber(destination, source, 0)
	}

	return OrRegisterRegister(destination, ZR, source)
}

// MoveRegisterNumber moves a number into the given register.
func MoveRegisterNumber(destination cpu.Register, number int) (uint32, bool) {
	if number < 0 {
		return MoveInvertedWideImmediate(destination, ^number, 0), true
	}

	if sizeof.Signed(number) <= 2 {
		return MoveZero(destination, 0, uint16(number)), true
	}

	if (number&0xFFFFFFFFFFFF == 0xFFFFFFFFFFFF) && sizeof.Signed(number>>48) <= 2 {
		return MoveInvertedWideImmediate(destination, (^number)>>48, 3), true
	}

	code, encodable := MoveBitmaskImmediate(destination, number)

	if encodable {
		return code, true
	}

	if (number&0xFFFFFFFF == 0xFFFFFFFF) && sizeof.Signed(number>>32) <= 2 {
		return MoveInvertedWideImmediate(destination, (^number)>>32, 2), true
	}

	if (number&0xFFFF == 0xFFFF) && sizeof.Signed(number>>16) <= 2 {
		return MoveInvertedWideImmediate(destination, (^number)>>16, 1), true
	}

	return 0, false
}

// MoveBitmaskImmediate moves a bitmask immediate value to a register.
func MoveBitmaskImmediate(destination cpu.Register, number int) (uint32, bool) {
	return OrRegisterNumber(destination, ZR, number)
}

// MoveInvertedWideImmediate moves an inverted 16-bit immediate value to a register.
func MoveInvertedWideImmediate(destination cpu.Register, number int, shift uint32) uint32 {
	return 0b100100101<<23 | shift<<21 | regImm(destination, number)
}

// MoveKeep moves a 16-bit integer into the given register and keeps all other bits.
func MoveKeep(destination cpu.Register, halfword int, number uint16) uint32 {
	return mov(0b11, halfword, number, destination)
}

// MoveZero moves a 16-bit integer into the given register and clears all other bits to zero.
func MoveZero(destination cpu.Register, halfword int, number uint16) uint32 {
	return mov(0b10, halfword, number, destination)
}

// mov encodes a generic move instruction.
func mov(opCode uint32, halfword int, number uint16, destination cpu.Register) uint32 {
	return 1<<31 | opCode<<29 | 0b100101<<23 | uint32(halfword<<21) | uint32(number<<5) | uint32(destination)
}
